@@ -1,7 +1,7 @@
 import pickle as pickle
-from typing import Any
+from typing import Any, Iterable
 from pydantic import BaseModel
-from typing import Dict, Union, List, Tuple
+from typing import Dict, Union, List, Tuple, Callable, Sized
 from datetime import datetime
 import pytz
 from dateutil import parser
@@ -9,7 +9,7 @@ import pandas as pd
 import pathlib
 import numpy as np
 from tqdm.auto import tqdm
-from multiprocessing import cpu_count
+from multiprocessing import cpu_count, Pool
 
 
 class PickleUtils(BaseModel):
@@ -129,7 +129,7 @@ def human_time(timestamp_sec: Union[float, str, int], formatter: str = '%Y-%m-%d
     # If input is a string, try to parse it as a float or int
     if isinstance(timestamp_sec, str):
         if timestamp_sec.replace('.', '', 1).isdigit():
-            timestamp_sec: float = int(timestamp_sec)
+            timestamp_sec: float = float(timestamp_sec)
         else:
             raise ValueError(f"Could not interpret string as a numeric timestamp: {timestamp_sec})")
     datetime_string: str = datetime.fromtimestamp(timestamp_sec, pytz.timezone(timezone)).strftime(formatter)
@@ -286,6 +286,65 @@ def divvy_workload(num_workers: int, tasks: List[Any]) -> List[List[Any]]:
     return task_list_all
 
 
+def multiprocess(func: Callable, iterable: List[Any], pool_function: str = None, batching: bool = False,
+                 num_workers: int = None, cuts_per_worker: int = 3, **kwargs) -> List[Any]:
+    """
+    Convenience wrapper for Pool.map and Pool.starmap that offers manual batching and automatic flattening
+
+    :param func: function to be evaluated in parallel
+    :param iterable: can be map or starmap style iterable of inputs to func
+    :param pool_function: whether to use 'map' or 'starmap'. Will try to infer if not provided.
+    :param batching: whether to handle batching manually
+    :param num_workers: number of processes to run in parallel (should not exceed CPU core count)
+    :param cuts_per_worker: if batching manually, can break into extra chunks to avoid idle time even if runtimes vary
+    :param kwargs: additional keyword arguments for 'map' or 'starmap' (for example the max chunk size)
+    :return: the results of func evaluated over the iterable
+    """
+    # Get num workers if not supplied
+    if not num_workers:
+        num_workers = get_num_workers(num_workers)
+
+    # Attempt to infer the pool function to be used, if not specified
+    # This is NOT infallible! Suppose you want to 'map' over a list of tuples; it would infer that 'starmap' is desired.
+    if not pool_function:
+        if isinstance(iterable[0], Iterable) and (not isinstance(iterable[0], str)):
+            pool_function: str = 'starmap'
+        else:
+            pool_function: str = 'map'
+
+    # Break the task list into batches if desired
+    if batching:
+        global batched_func  # ... TODO: how to make this pickleable without making it global??
+
+        def batched_func(tasks: List[Any]) -> List[Any]:  # noqa:
+            return [func(args) for args in tasks]
+
+        if pool_function == 'starmap':
+            raise NotImplementedError(f"batching and starmap are mutually exclusive in the current implementation")
+
+        map_style_inputs: List[List[Any]] = divvy_workload(num_workers=num_workers * cuts_per_worker, tasks=iterable)
+        use_function: Callable = batched_func
+    else:
+        # No batching = batches of one (let the pool methods handle chunksize)
+        map_style_inputs: List[Any] = iterable
+        use_function = func
+
+    # Run the pool
+    with Pool(num_workers) as pool:
+        if pool_function.lower() == 'starmap':
+            result = pool.starmap(func=use_function, iterable=map_style_inputs, **kwargs)
+        elif pool_function.lower() == 'map':
+            result = pool.map(func=use_function, iterable=map_style_inputs, **kwargs)
+        else:
+            raise ValueError(f"Pool function should be 'map' or 'starmap' but received unknown type: {pool_function}")
+
+    # Flatten if necessary
+    if batching:
+        result = [item for sublist in result for item in sublist]
+
+    return result  # noqa: results prouced in lines above
+
+
 def object_string_merge(string: str, values_from: Any, left_merge_token: str = '[[',
                         right_merge_token: str = ']]') -> str:
     """
@@ -343,10 +402,19 @@ def risky_cast(x: Any) -> Any:
             except:
                 pass
 
-    # Bool?
-    str_to_bool_mapper: Dict[str, Any] = {'true': True, 'false': False}
-    if x.lower() in str_to_bool_mapper:
-        return str_to_bool_mapper[x.lower()]
+    # Bool? None?
+    mapper: Dict[str, Any] = {'true': True, 'false': False, 'none': None}
+    if x.lower() in mapper:
+        return mapper[x.lower()]
 
     # Give up
     return x
+
+
+def flatten_list(nested_list: List[List[Any]]) -> List[Any]:
+    """
+    Helper function that flattens a list, nothing fancy (only works 1 level deep)
+    :param nested_list: a list of lists of <something>
+    :return: a list of <something>
+    """
+    return [item for sublist in nested_list for item in sublist]
